# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import uuid
from http import HTTPStatus
from typing import Any
from typing import Dict

import requests
from requests import HTTPError
from requests_toolbelt import user_agent
from requests_toolbelt.adapters import socket_options
from urllib3.util import Retry

from kolena import __name__ as client_name
from kolena import __version__ as client_version
from kolena._utils.endpoints import get_endpoint
from kolena._utils.state import DEFAULT_API_VERSION
from kolena._utils.state import get_client_state
from kolena._utils.state import kolena_initialized
from kolena.errors import IncorrectUsageError
from kolena.errors import NameConflictError
from kolena.errors import NotFoundError
from kolena.errors import RemoteError
from kolena.errors import UnauthenticatedError

__all__ = [
    "get",
    "post",
    "put",
    "delete",
    "raise_for_status",
]

# Give the client 15 seconds to connect to kolena server
# Slightly more than a multiple of 3, as per https://docs.python-requests.org/en/master/user/advanced/#timeouts
CONNECTION_CONNECT_TIMEOUT = 15.05
CONNECTION_READ_TIMEOUT = 60 * 60  # Give kolena server 1 hour to respond to client request

# This only retries for failed DNS lookups, socket connections and connection timeouts.
# HTTPAdapter sets this to 0 by default. https://requests.readthedocs.io/en/latest/_modules/requests/adapters/
# Using the Retry object to configure a backoff which is not supported by using an int here.
MAX_RETRIES = Retry(total=3, connect=3, read=0, redirect=0, status=0, backoff_factor=2)

# This retries for read errors in addition to the ones mentioned above.
# Should only be used for idempotent operations.
MAX_IDEMPOTENT_RETRIES = Retry(total=3, connect=3, read=3, redirect=0, status=0, backoff_factor=2)


class JWTAuth(requests.auth.AuthBase):
    """Attaches JWT Authorization to the given Request object"""

    def __init__(self, jwt: str):
        self.jwt = jwt

    def __call__(self, r: Any) -> Any:
        r.headers["Authorization"] = f"Bearer {self.jwt}"
        return r


@kolena_initialized
def _with_default_kwargs(**kwargs: Any) -> Dict[str, Any]:
    client_state = get_client_state()
    assert client_state.jwt_token is not None
    default_kwargs = {
        "auth": JWTAuth(client_state.jwt_token),
        "timeout": (CONNECTION_CONNECT_TIMEOUT, CONNECTION_READ_TIMEOUT),
        "proxies": client_state.proxies,
    }
    default_headers = client_state.additional_request_headers or {}
    default_headers.update(
        {
            "Content-Type": "application/json",
            "X-Request-ID": uuid.uuid4().hex,
            "User-Agent": user_agent(client_name, client_version),
        },
    )
    # allow requests to override Content-Type as needed by for example file uploads
    kw_headers = kwargs.get("headers", {})
    if "Content-Type" in kw_headers:
        default_headers["Content-Type"] = kw_headers["Content-Type"]
    return {
        **kwargs,
        **default_kwargs,
        "headers": {**kwargs.get("headers", {}), **default_headers},
    }


@kolena_initialized
def get(
    endpoint_path: str,
    params: Any = None,
    api_version: str = DEFAULT_API_VERSION,
    **kwargs: Any,
) -> requests.Response:
    url = get_endpoint(endpoint_path=endpoint_path, api_version=api_version)
    with requests.Session() as s:
        s.mount("https://", socket_options.TCPKeepAliveAdapter(max_retries=MAX_IDEMPOTENT_RETRIES))
        return s.get(url=url, params=params, **_with_default_kwargs(**kwargs))


@kolena_initialized
def post(
    endpoint_path: str,
    data: Any = None,
    json: Any = None,
    api_version: str = DEFAULT_API_VERSION,
    **kwargs: Any,
) -> requests.Response:
    url = get_endpoint(endpoint_path=endpoint_path, api_version=api_version)
    with requests.Session() as s:
        s.mount("https://", socket_options.TCPKeepAliveAdapter(max_retries=MAX_RETRIES))
        return s.post(url=url, data=data, json=json, **_with_default_kwargs(**kwargs))


@kolena_initialized
def put(
    endpoint_path: str,
    data: Any = None,
    json: Any = None,
    api_version: str = DEFAULT_API_VERSION,
    **kwargs: Any,
) -> requests.Response:
    url = get_endpoint(endpoint_path=endpoint_path, api_version=api_version)
    with requests.Session() as s:
        s.mount("https://", socket_options.TCPKeepAliveAdapter(max_retries=MAX_IDEMPOTENT_RETRIES))
        return s.put(url=url, data=data, json=json, **_with_default_kwargs(**kwargs))


@kolena_initialized
def delete(endpoint_path: str, api_version: str = DEFAULT_API_VERSION, **kwargs: Any) -> requests.Response:
    url = get_endpoint(endpoint_path=endpoint_path, api_version=api_version)
    with requests.Session() as s:
        s.mount("https://", socket_options.TCPKeepAliveAdapter(max_retries=MAX_IDEMPOTENT_RETRIES))
        return requests.delete(url=url, **_with_default_kwargs(**kwargs))


def raise_for_status(response: requests.Response) -> None:
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        # HTTP 401 is "unauthorized" but used as "unauthenticated"
        raise UnauthenticatedError(response.content)
    if response.status_code == HTTPStatus.NOT_FOUND:
        raise NotFoundError(response.content)
    if response.status_code == HTTPStatus.CONFLICT:
        raise NameConflictError(response.content)
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        raise IncorrectUsageError(response.content)

    try:
        response.raise_for_status()
    except HTTPError:
        raise RemoteError(f"{response.text} ({response.elapsed.total_seconds():0.5f} seconds elapsed)")


@kolena_initialized
def get_connection_args(**kwargs: Any) -> Dict[str, Dict[str, str]]:
    client_state = get_client_state()
    return {"proxies": client_state.proxies}



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List
from typing import Tuple

from shapely.geometry import Polygon
from shapely.validation import make_valid

from kolena._utils.pydantic_v1 import validate_arguments
from kolena._utils.validators import ValidatorConfig


def make_valid_polygon(points: List[Tuple[float, float]]) -> None:
    return make_valid(Polygon(points))


@validate_arguments(config=ValidatorConfig)
def validate_polygon(points: List[Tuple[float, float]]) -> None:
    try:
        make_valid_polygon(points)
    except Exception as exception:
        raise ValueError("Point set is an invalid polygon", points) from exception



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
from typing import List
from typing import Union

from kolena._experimental.object_detection.workflow import TestSample as BaseTestSample
from kolena._experimental.object_detection.workflow import ThresholdConfiguration
from kolena._utils.pydantic_v1.dataclasses import dataclass
from kolena.workflow import define_workflow
from kolena.workflow import GroundTruth as BaseGroundTruth
from kolena.workflow import Inference as BaseInference
from kolena.workflow.annotation import LabeledBoundingBox
from kolena.workflow.annotation import LabeledPolygon
from kolena.workflow.annotation import ScoredLabeledPolygon


@dataclass(frozen=True)
class TestSample(BaseTestSample):
    """The [`Image`][kolena.workflow.Image] sample type for the pre-built 2D Instance Segmentation workflow."""


@dataclass(frozen=True)
class GroundTruth(BaseGroundTruth):
    """Ground truth type for the pre-built 2D Instance Segmentation workflow."""

    polygons: List[LabeledPolygon]
    """
    The ground truth [`LabeledPolygons`][kolena.workflow.annotation.LabeledPolygon] associated with an image.
    """

    ignored_polygons: List[Union[LabeledBoundingBox, LabeledPolygon]] = dataclasses.field(default_factory=list)
    """
    The ground truth [`LabeledPolygons`][kolena.workflow.annotation.LabeledPolygon] to be ignored
    in evaluation associated with an image.
    """

    n_polygons: int = dataclasses.field(default_factory=lambda: 0)

    def __post_init__(self):
        object.__setattr__(self, "n_polygons", len(self.polygons))


@dataclass(frozen=True)
class Inference(BaseInference):
    """Inference type for the pre-built 2D Instance Segmentation workflow."""

    polygons: List[ScoredLabeledPolygon]
    """
    The inference [`ScoredLabeledPolygons`][kolena.workflow.annotation.ScoredLabeledPolygon] associated with an image.
    """

    ignored: bool = False
    """
    Whether the image (and its associated inference `polygons`) should be ignored
    in evaluating the results of the model.
    """


_, TestCase, TestSuite, Model = define_workflow(
    "Instance Segmentation",
    TestSample,
    GroundTruth,
    Inference,
)


@dataclass(frozen=True)
class EvaluatorConfiguration(ThresholdConfiguration):
    """
    Confidence and [IoU ↗](../../metrics/iou.md) threshold configuration for the pre-built
    2D Instance Segmentation workflow.
    Specify a confidence and IoU threshold to apply to all classes.
    """



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set

from kolena._api.v1.batched_load import BatchedLoad
from kolena._utils.pydantic_v1.dataclasses import dataclass


class Model:
    @dataclass(frozen=True)
    class CreateRequest:
        name: str
        metadata: Dict[str, Any]
        workflow: str
        tags: Optional[List[str]] = None

    @dataclass(frozen=True)
    class LoadByNameRequest:
        name: str

    @dataclass(frozen=True)
    class LoadAllRequest:
        workflow: str
        tags: Optional[List[str]] = None

    @dataclass(frozen=True)
    class EntityData:
        id: int
        name: str
        metadata: Dict[str, Any]
        tags: Set[str]
        workflow: str

    @dataclass(frozen=True)
    class LoadAllResponse:
        models: List["Model.EntityData"]

    @dataclass(frozen=True)
    class DeleteRequest:
        id: int


@dataclass(frozen=True)
class TestCaseInfo:
    id: int
    name: str
    sample_count: int
    # provided by stratification preview to show what bucket this test-case falls into,
    # e.g. {"count": "low", "gender": "female"}. This info is not available from stratification result.
    membership: Optional[Dict[str, str]] = None


class BulkProcessStatus(Enum):
    CREATED = "created"
    LOADED = "loaded"
    EDITED = "edited"


class TestCase:
    @dataclass(frozen=True)
    class CreateRequest:
        name: str
        description: str
        workflow: str

    @dataclass(frozen=True)
    class SingleProcessRequest:
        name: str
        reset: bool = False

    @dataclass(frozen=True)
    class SingleProcessResponse:
        data: "TestCase.EntityData"
        status: BulkProcessStatus

    @dataclass(frozen=True)
    class BulkProcessRequest:
        test_cases: List["TestCase.SingleProcessRequest"]
        workflow: str
        uuid: Optional[str] = None

    @dataclass(frozen=True)
    class BulkProcessResponse:
        test_cases: List["TestCase.SingleProcessResponse"]

    @dataclass(frozen=True)
    class CreateFromExistingRequest(BatchedLoad.WithLoadUUID):
        test_case_name: str
        test_suite_name: str
        workflow: str
        compute_metrics_where_possible: bool = False
        compute_metrics_models: Optional[List[int]] = None

    @dataclass(frozen=True)
    class CreateFromExistingResponse:
        test_case_id: int
        test_suite_id: int  # newly created version ID

    @dataclass(frozen=True)
    class LoadByNameRequest:
        name: str
        version: Optional[int] = None

    @dataclass(frozen=True)
    class EntityData:
        id: int
        name: str
        version: int
        description: str
        workflow: str

    @dataclass(frozen=True)
    class LoadContentsRequest:
        test_case_id: int

    @dataclass(frozen=True)
    class InitLoadContentsRequest(LoadContentsRequest, BatchedLoad.BaseInitDownloadRequest):
        ...

    @dataclass(frozen=True)
    class EditRequest:
        test_case_id: int
        current_version: int
        description: str
        reset: bool = False

    @dataclass(frozen=True)
    class CompleteEditRequest(EditRequest, BatchedLoad.WithLoadUUID):
        ...


class TestSuite:
    @dataclass(frozen=True)
    class CreateRequest:
        name: str
        description: str
        workflow: str
        tags: Optional[List[str]] = None

    @dataclass(frozen=True)
    class LoadByNameRequest:
        name: str
        version: Optional[int] = None

    @dataclass(frozen=True)
    class LoadAllRequest:
        workflow: str
        tags: Optional[List[str]] = None

    @dataclass(frozen=True)
    class EntityData:
        id: int
        name: str
        version: int
        description: str
        test_cases: List[TestCase.EntityData]
        workflow: str
        tags: List[str]

    @dataclass(frozen=True)
    class LoadAllResponse:
        test_suites: List["TestSuite.EntityData"]

    @dataclass(frozen=True)
    class EditRequest:
        test_suite_id: int  # ID of version being edited
        current_version: int
        name: str
        description: str
        test_case_ids: List[int]
        tags: Optional[List[str]] = None  # unique set -- list used to preserve ordering

    @dataclass(frozen=True)
    class DeleteRequest:
        test_suite_id: int


class TestRun:
    @dataclass(frozen=True)
    class MarkCrashedRequest:
        test_run_id: int


Model.LoadAllResponse.__pydantic_model__.update_forward_refs()  # type: ignore
TestCase.SingleProcessResponse.__pydantic_model__.update_forward_refs()  # type: ignore[attr-defined]
TestCase.BulkProcessRequest.__pydantic_model__.update_forward_refs()  # type: ignore[attr-defined]
TestCase.BulkProcessResponse.__pydantic_model__.update_forward_refs()  # type: ignore[attr-defined]
TestSuite.LoadAllResponse.__pydantic_model__.update_forward_refs()  # type: ignore
